(*
 *
 * Copyright (c) 2004, 
 *  Ryan Cunningham    <rcunning@cs.ucla.edu>
 * All rights reserved.
 *
 *)
(* eelstatechart.ml *)
(* code for eelstatechart.mli *)

(* Open some other cil modules *)
open Cil
open Trace
open Printf
module P = Pretty

(* -------------------------------------------------------------------------- *)
(* Begin State Chart generation types, functions, and classes                 *)

(* --- define types --- *)

type mutable_int = {
  mutable value_int:int;
}

type mutable_bool = {
  mutable value:bool;
}

(* static callgraph node *)
type eelcallnode = {
  (* this varinfo is for the fundec *)
  ecnInfo: varinfo;
  
  (* Calls: function name, queue of calls (call node * location of call 
   *        * fundec of called function * call stmt * call instr * call exp list) *)
  ecnCalls: 
    (string, (eelcallnode * location * fundec * stmt * instr * exp list) Queue.t) Hashtbl.t;
    
  (* Callers: calling function name, queue of callers (callers node * location of call 
   *          * fundec of calling function * calling stmt * calling instr * calling exp list) *)
  ecnCallers: 
    (string, (eelcallnode * location * fundec * stmt * instr * exp list) Queue.t) Hashtbl.t;
  
  (* Callbacks: callback name, queue of (callback node * location of the call that generated 
   *            the callback *)
  ecnCallbacks: (string, (eelcallnode * location) Queue.t) Hashtbl.t;
  
  (* Calledbacks: callback name, queue of (callback node * location of the call that generated
   *              the callback *)
  ecnCalledbacks: (string, (eelcallnode * location) Queue.t) Hashtbl.t;
  
  (* This is a hashtable that hold more detailed information about callbacks generated
   * See eelTypeHash below for more information *)
  ecnType: eelTypeHash;
  
  (* boolean to signify if this call node is a callback leaf (i.e. generates a callback through
   * an actual libeel call) *)
  mutable ecnLeaf: bool; 
}

(* libeel type hash: name of libeel function * libeel function's node
 * * location of the libeel call * group id source
 * * hash of operation performed on the group id *)
and eelTypeHash =
  (string * eelcallnode * location * eelgid_source, (eelgid_operation, unit) Hashtbl.t) Hashtbl.t

(* group id source (i.e. source has been found) *)  
and eelgid_source = 
  | SrcGlobal   of varinfo             (* global group id variable with varinfo *)
  | SrcArgument of int                 (* argument number to calling function *)
  | SrcNew      of varinfo * location  (* new group id: the group id's varinfo, and location of call to eel_new_group_id *)
  | SrcUnknown                         (* unknown source *)
  | SrcNotVar                          (* the group id source is not a variable (unknown) *)
  | SrcCall                            (* the group id source is from the return of a function call (unknown) *)
  
(* group id operation *)
and eelgid_operation = 
  | AddWrite of location * string * eelgid_arg   (* state name * gid argument *)
  | AddRead  of location * string * eelgid_arg   (* state name * gid argument *)
  | AddError of location * string * eelgid_arg   (* state name * gid argument *)
  | AddTimer of location * string * eelgid_arg   (* state name * gid argument *)
  | AddTimeout of location * string * eelgid_arg (* state name * gid argument *)
  | DeleteGid of location * eelgid_arg           (* gid argument *)
  | NewGid   of location * eelgid_arg            (* new gid var *)

(* group id argument information (i.e. before actual source is found) *)
and eelgid_arg = 
  | ArgGlobal   of varinfo        (* argument is a global variable *)
  | ArgLocal    of varinfo        (* argument is a local variable *)
  | ArgArgument of varinfo * int  (* argument is an argument to the calling function *)
  | ArgUnknown                    (* argument is something else *)

(* top level type for the static callgraph: name of function, its node *)
type eelcallgraph = 
  (string, eelcallnode) Hashtbl.t

(* top level type for the "states" = all callbacks in the program *)
type eelstates = 
  (string, eelcallnode) Hashtbl.t


(* --- implementation --- *)

(* Some general utility helpers *)

(* add something to a hashtable if it's not already there, instead of
 * the default pushback behavior of Hashtbl.add *)
let add_if (table : ('a,'b) Hashtbl.t) (key:'a) (data:'b) : unit =
begin
  if (not (Hashtbl.mem table key)) then
    (Hashtbl.add table key data)
end

(* add something to the queue of a hashtable's data *)
let add_to_HtoQ (table : ('a,'b Queue.t) Hashtbl.t) 
                (key:'a) 
                (data:'b) : unit =
begin
  try
    let dataQ = (Hashtbl.find table key) in
    (Queue.add data dataQ);
  with Not_found -> (
    let newQ:'b Queue.t = (Queue.create ()) in
    (Queue.add data newQ);
    (Hashtbl.add table key newQ); 
  )
end

(* add something to the hashtbale of a hasthable's data *)
let add_to_HtoH (table : ('a,('b, 'c) Hashtbl.t) Hashtbl.t) 
                (key1:'a) 
                (key2:'b)
                (data:'c) : unit =
begin
  try
    let dataH = (Hashtbl.find table key1) in
    (add_if dataH key2 data);
  with Not_found -> (
    let newH:('b, 'c) Hashtbl.t = (Hashtbl.create 1) in
    (Hashtbl.add newH key2 data);
    (Hashtbl.add table key1 newH); 
  )
end

(* add something to the queue of a hashtable's data if key is
 * not already present *)
let add_if_HtoQ (table : ('a,'b Queue.t) Hashtbl.t) 
                (key:'a) 
                (data:'b) : unit =
begin
  try
    ignore (Hashtbl.find table key);
  with Not_found -> (
    let newQ:'b Queue.t = (Queue.create ()) in
    (Queue.add data newQ);
    (Hashtbl.add table key newQ); 
  )
end

(* iterate through all items of a hashtable that has a queue as its data *)
let iterHtoQ (func) (h:('a, 'b Queue.t) Hashtbl.t) : unit = 
begin
  let hIter (key:'a) (data:'b Queue.t) : unit = 
    let qIter (data_inner:'b) : unit = 
      (func key data_inner) in
    (Queue.iter qIter data) in
  (Hashtbl.iter hIter h);
end

(* iterate through all items of a hashtable that has a hashtable as its data *)
let iterHtoH (func) (h:('a, ('b, 'c) Hashtbl.t) Hashtbl.t) : unit = 
begin
  let h1Iter (key1:'a) (h2:('b, 'c) Hashtbl.t) : unit = 
    let h2Iter (key2:'b) (data:'c) : unit = 
      (func key1 key2 data) in
    (Hashtbl.iter h2Iter h2) in
  (Hashtbl.iter h1Iter h);
end

(* fold a hashtable with hashtable as data to a list of triples *)
let foldHtoH (h:('a, ('b, 'c) Hashtbl.t) Hashtbl.t) : ('a * 'b * 'c) list = 
  let h1Iter (key1:'a) (h2:('b, 'c) Hashtbl.t) (l:('a * 'b * 'c) list) : ('a * 'b * 'c) list = 
    let h2Iter (key2:'b) (data:'c) (li:('a * 'b * 'c) list) : ('a * 'b * 'c) list = 
      (key1,key2,data) :: li in
    (Hashtbl.fold h2Iter h2 l) in
  (Hashtbl.fold h1Iter h [])

(* iterate through all matching keys of two hashtables that have queues as data,
 * call func for each possible pair of queue items *)
let iterMatchkeysHtoQ (func) 
                      (h1:('a, 'b Queue.t) Hashtbl.t)
                      (h2:('a, 'c Queue.t) Hashtbl.t) : unit =
begin
  let matchedKeyIter (func) (key:'a) (q1:'b Queue.t) (q2:'c Queue.t) : unit =
  begin 
    let q1Iter (data1:'b) : unit = 
      let q2Iter (data2:'c) : unit = 
        (func key data1 data2) in
      (Queue.iter q2Iter q2) in
    (Queue.iter q1Iter q1);
  end in
  
  let h1Iter (key:'a) (h1Data:'b Queue.t) : unit = 
  begin
    try
      let h2Data:'c Queue.t = (Hashtbl.find h2 key) in
      (matchedKeyIter func key h1Data h2Data);
    with Not_found -> ();
  end in
  
  (Hashtbl.iter h1Iter h1);
end

(* convert a Hashtbl to a List *)
let hashToList (table:('a, 'b) Hashtbl.t) : ('a * 'b) list =
begin
  let foldFunc 
          (key:'a) 
          (value:'b) 
          (l:('a * 'b) list) : ('a * 'b) list =
    (key, value) :: l in
  (Hashtbl.fold foldFunc table [])
end

(* sort by the first item of a list of pairs *)
let sortByFirst (l:('a * 'b) list) : ('a * 'b) list = 
begin
  let compareFirst (item1: 'a * 'b) (item2: 'a * 'b) : int =
    (compare (fst item1) (fst item2)) in
  (List.sort compareFirst l)
end

(* iterate on a Hashtbl in order of sorted keys *)
let sortByKeyAndIter (iterFunc) (table: ('a, 'b) Hashtbl.t) : unit = 
begin
  let funcProxy (item: 'a * 'b) : unit =
    (iterFunc (fst item) (snd item)) in
  (List.iter funcProxy 
    (sortByFirst 
      (hashToList table)))
end 

(* Helper to ensure that all items in a list match using =, if not return default_ret *)
let matchAllElse (l:'a list) (default_ret:'a) : 'a =
begin
  let rec matchAllElseWith (prev_item:'a) (rest:'a list) : 'a =
    (match rest with
    | [] -> prev_item
    | hd :: tl when hd = prev_item -> (matchAllElseWith hd tl)
    | _ -> default_ret) in
  (match l with
  | hd :: tl -> (matchAllElseWith hd tl)
  | _        -> default_ret)
end

(* Some eelstatechart specific helpers *)

(* iterate through an eelTypeHash - placing calls to eel_new_group_id first *)
let iterTypeByNews (func) (type_hash:eelTypeHash) : unit =
begin
  let compareQuadTriple 
    (x:((string * eelcallnode * Cil.location * eelgid_source) * eelgid_operation * unit)) 
    (y:((string * eelcallnode * Cil.location * eelgid_source) * eelgid_operation * unit)) : int =
    (match x, y with (xx,_,_), (yy,_,_) -> 
      (match xx, yy with 
      | (_,_,_,SrcNew(_,_)), (_,_,_,SrcNew(_,_)) -> 0
      | (_,_,_,_), (_,_,_,SrcNew(_,_)) -> 1
      | (_,_,_,SrcNew(_,_)), (_,_,_,_) -> -1
      | _,_ -> 0
      )
    ) in
  let iterSorted (tr:
    ((string * eelcallnode * Cil.location * eelgid_source) * eelgid_operation * unit)) : unit =
    (match tr with ft,st,tt -> (func ft st tt)) in
  (List.iter iterSorted (List.sort compareQuadTriple (foldHtoH type_hash)));
end
    
(* Some cil only specific helpers (NOT USED) *)

(* get the name of an instr that is a Call *)
(* let get_callName (i:instr) : string =
    (match i with
    | Call(_,Lval(Var(vi),NoOffset),_,_) -> vi.vname
    | _ -> raise Not_found) *)

(* helper for stmtReachable *)
(* let rec stmtReachableHelper (sfrom:stmt) (sto:stmt) (seen:(int, bool) Hashtbl.t) : bool = 
begin
  if (not (Hashtbl.mem seen sfrom.sid)) then
  begin
    (Hashtbl.add seen sfrom.sid true);
    if sfrom.sid = sto.sid then 
      true 
    else
    begin
      let reachable:mutable_bool = {value=false} in
      let inListDFS (s_inner:stmt) : unit = 
        if (not reachable.value) then 
          if (stmtReachableHelper s_inner sto seen) then
            reachable.value <- true in
      (List.iter inListDFS sfrom.succs);
      reachable.value
    end
  end
  else
    false
end *)

(* return true if sfrom is reachable from sto
 * the function both are in must be CFGed before
 * using this *)
(* let stmtReachable (sfrom:stmt) (sto:stmt) : bool = 
  (stmtReachableHelper sfrom sto (Hashtbl.create 10)) *)

(* helper for getReturnStmts *)
(* let rec getReturnStmtsSeen (s:stmt) (seen:(int,bool) Hashtbl.t) : stmt list = 
begin
  if (Hashtbl.mem seen s.sid) then [] else 
  begin
    (Hashtbl.add seen s.sid true);
    let rec getSuccsReturns (sl:stmt list) : stmt list =
    begin
      match sl with
      | hd :: tl -> (List.append (getReturnStmtsSeen hd seen) (getSuccsReturns tl));
      | [] -> [];
    end in
    match s.skind with
    | Return(e,loc) -> s :: (getSuccsReturns s.succs);
    | _ -> (getSuccsReturns s.succs);
  end
end *)
(* get the possible return stmts starting from stmt s,
 * function (fundec) must be CFGed first using prepareCFG 
 * and computeCFGInfo *)
(* let getReturnStmts (s:stmt) : stmt list = 
  (getReturnStmtsSeen s (Hashtbl.create 15)) *)
  
(* find the index of a particular vi in a vi list *)
let rec findVi (vi:varinfo) (vl:varinfo list) (cur:int) : int = 
  (match vl with
  | [] -> -1
  | hd :: _ when hd.vid = vi.vid -> cur
  | _ :: tl -> (findVi vi tl (cur + 1)))
  
  
(* Classes *)
    
(* a cil visitor class implementation that builds the
 * callgraph and callback graph portion of the state 
 * chart *)
class eelscBuildCallgraph = object(self)
  inherit nopCilVisitor

  (* this is the graph we are computing,
   * it is created empty when the object is constructed, and will
   * be added-to during the computation *)
  val graph: eelcallgraph = (Hashtbl.create 117)
  
  (* these are the states of the application (i.e. callback functions) *)
  val states: eelstates = (Hashtbl.create 10)
  
  (* cache a sid * vid hash of gid sources *)
  val varSourceCache: (int * int, eelgid_source) Hashtbl.t = (Hashtbl.create 20)

  (* the current function we're in, so when we visit a call node
   * we know who is the caller, save these between different visitor method invocations *)
  val mutable curFunc: eelcallnode option = None
  val mutable curFundec: fundec option = None
  val mutable curStmt: stmt option = None
  val mutable curInstr: instr option = None

  (* return the computed call graph *)
  method getGraph () : eelcallgraph = graph
  (* return the computed states *)
  method getStates () : eelstates = states
  
  (* given the name of a function, retrieve its eelcallnode; this
   * will create a node if one doesn't already exist *)
  method getNode (info:varinfo) : eelcallnode =
  begin
    let name = info.vname in
    try
      (Hashtbl.find graph name)
    with Not_found -> (
      (* make a new node *)
      let ret:eelcallnode = {
        ecnInfo = info;
        ecnCalls = (Hashtbl.create 5);
        ecnCallers = (Hashtbl.create 5);
        ecnCallbacks = (Hashtbl.create 5);
        ecnCalledbacks = (Hashtbl.create 5);
        ecnType = (Hashtbl.create 0);
        ecnLeaf = false;
      }  in
      (* add it to the graph table, then return it *)
      (Hashtbl.add graph name ret);
      ret
    )
  end
  
  (* wrap the potential vi with an argument classification (Global/Local/Unknown) *)
  method getGidArg (vi_op:varinfo option) : eelgid_arg =
  begin
    match vi_op, curFundec, curStmt with
    | Some(vi), Some(f), Some(s) -> (
      (* if it is a global, just return such *)
      if vi.vglob then (ArgGlobal vi) else
      (* find if it is an argument or a local *)
      let arg_num: int = (findVi vi f.sformals 0) in 
      if arg_num > -1 then 
        (ArgArgument (vi, arg_num))
      else if (List.exists (fun v -> v.vid = vi.vid) f.slocals) then 
        (ArgLocal vi)
      else
        ArgUnknown)
    | _ -> ArgUnknown;
  end
  
  (* get the vi from an exp, if it is actually a variable (or cast of a variable) *)
  method getGidVi (e:exp) : varinfo option =
  begin
    (match e with
    | Lval(l:Cil.lval) -> 
        (match l with
        | (Cil.Var(vi),_) -> (Some vi)
        | (Cil.Mem(_),_) -> None)
    | CastE(_,e) -> (self#getGidVi e)
    | _ -> None); (* some random exp that will not end up being an eel call *)
  end
  
  (* try to get the 0th vi from an exp list *)
  method getGidViExpList (explist:exp list) : varinfo option =
  begin
    try
      (self#getGidVi (List.nth explist 0));
    with _ -> None (* nth failed *)
  end
  
  (* try to get the nth vi from an exp list *)
  method getGidViExpNth (explist:exp list) (n:int) : varinfo option =
  begin
    try
      (self#getGidVi (List.nth explist n));
    with _ -> None (* nth failed *)
  end
  
  (* pull out the varinfo of a call argument, or None if it is not there *)
  method getCallbackVi (explist:exp list)
                       (argument_number:int)
                       (loc:location) : varinfo option =
  begin
    try
      (* only add it if the argument is a named function (not pointer) *)
      (match (List.nth explist argument_number) with
      | AddrOf(l:Cil.lval) -> 
          (match l with
          | (Cil.Var(vi),NoOffset) -> (Some vi)
          | _ -> None);  (* this is an offset into a struct *)
      | _ -> None);      (* not an lval, i.e. not a named function *)
    with _ -> None (* nth failed *)
  end
  
  (* pseudo-logic for finding gid source:
   *  Unknown -> Unknown
   *  Global -> Global
   *  Arg -> Arg if no take addr of or set in func, otherwise Unknown
   *  Local -> look for first set from new gid call, all preds must reach the same set, otherwise Unknown
   *)
  
  (* get the actual gid source from the call arguement information *)
  method getGidSource (gid_arg:eelgid_arg) : eelgid_source =
  begin

    (* does the fundec ever asign the variable *)
    let usesArg (f:fundec) (vii:varinfo) : bool =
    begin
      if (vii.vaddrof) then 
        true 
      else
        (
        let uses = ref false in
        let iIter (i:instr) : unit =
          (match i with
          | Set((Var(vic),NoOffset),_,_)
          | Call(Some(Var(vic),NoOffset),_,_,_) when vic.vid = vii.vid -> uses := true;
          | _ -> ()) in
        let sIter (s:stmt) : unit =
          (match s.skind with 
          | Instr(il) -> (List.iter iIter il)
          | _ -> ()) in
        (List.iter sIter f.sallstmts);
        !uses
        )
    end in
    
    (* get all sets/calls to do with this var within s, use preds to track back 
     * return the list of all these - caller should compare to see if they all match.
     * Note to Rupak: this should be fixed to merge variable states to prevent exponential 
     *                blow up *)
    let rec getPredSetInstrsSeen (s:stmt) (vid:int) (seen:(int,bool) Hashtbl.t) : eelgid_source list = 
    begin
      if (Hashtbl.mem seen s.sid) then [] else 
      begin
        (Hashtbl.add seen s.sid true);
        
        let rec getPredSets (sl:stmt list) (vidi:int) : eelgid_source list =
        begin
          match sl with
          | hd :: tl -> (List.append (getPredSetInstrsSeen hd vidi seen) (getPredSets tl vidi));
          | [] -> [];
        end in
        
        let rec getInstrSource (il:instr list) (vidi:int) : eelgid_source option * int =
        begin
          match il with
          | [] -> None, vidi
          | hd :: tl -> (
            (*(fprintf stdout "Looking at instr at %s:%d\n" (get_instrLoc hd).file (get_instrLoc hd).line);*)
            match hd with 
            | Set((Var(vic),NoOffset),Lval((Var(newvi),NoOffset)),_) when vic.vid = vidi -> 
              (* asigned from a different vi *)
              ((*(fprintf stdout "Switch to %s\n" newvi.vname);*)
              (getInstrSource tl newvi.vid))
            | Set((Var(vic),NoOffset),_,_) when vic.vid = vidi -> 
              (* asigned from something that is not a simple var...give up *)
              (Some (SrcNotVar)), 0
            | Call(Some(Var(vic),NoOffset),Lval((Var(newvi),NoOffset)),_,loc) 
                when vic.vid = vidi && newvi.vname = "eel_new_group_id" -> 
              (* asigned from a call to new gid *)
              (Some (SrcNew (vic, loc))), 0
            | Call(Some(Var(vic),NoOffset),_,_,_) when vic.vid = vidi -> 
              (* asigned from some other call...give up *)
              (Some (SrcCall)), 0
            | _ -> (getInstrSource tl vidi))
        end in
        
        match s.skind with
        | Instr(il) -> (
          match (getInstrSource (List.rev il) vid) with
          | Some(src), _ -> [src]
          | None, new_vid -> (getPredSets s.preds new_vid)
          )
        | _ -> (getPredSets s.preds vid);
      end
    end in

    (* unpack the source info and perform logic listed at top *)
    (match curFundec, curStmt, curInstr with
     | Some(f), Some(s), Some(i) -> (
       match gid_arg with
       | ArgGlobal(vi)     -> (SrcGlobal vi)
       | ArgLocal(vi)      -> (matchAllElse (getPredSetInstrsSeen s vi.vid (Hashtbl.create 15)) SrcUnknown)
       | ArgArgument(vi,n) -> (if (usesArg f vi) then SrcUnknown else (SrcArgument n))
       | ArgUnknown        -> SrcUnknown
       )
     | _ -> SrcUnknown)
  end
  
  (* just pull out the eelgid_source from the eelgid_operation *)
  method getGidSourceOp (gid_op:eelgid_operation) : eelgid_source =
  begin
    (self#getGidSource 
      (match gid_op with
       | AddWrite(_,_,arg)
       | AddRead(_,_,arg)
       | AddError(_,_,arg)
       | AddTimer(_,_,arg)
       | AddTimeout(_,_,arg) -> arg
       | DeleteGid(_,arg)
       | NewGid(_,arg) -> arg))
  end
    
  (* add an eel callback by event type and argument number *)
  method addeelCallback (caller:eelcallnode)
                        (callee:eelcallnode)
                        (loc:location)
                        (callback_vi:varinfo option)
                        (gid_op:eelgid_operation) : unit =
  begin
    (match callback_vi with
    | Some(vi) -> (
      let callback:eelcallnode = (self#getNode vi) in 
          (* add the callback to the global states *)
          (add_if states callback.ecnInfo.vname callback);
          (* add the callback and caller *)
          (add_to_HtoQ caller.ecnCallbacks callback.ecnInfo.vname (callback, loc));
          (add_to_HtoQ callback.ecnCalledbacks caller.ecnInfo.vname (caller, loc));
          (* mark caller as generating a gid operation at this loc *)
          (add_to_HtoH caller.ecnType 
            (callee.ecnInfo.vname, callee, loc, (self#getGidSourceOp gid_op)) gid_op ());
          )
    | _ -> ());
  end
      
  (* add the eel calls for a valid function call *)
  method addeelCalls (caller:eelcallnode)
                     (callee:eelcallnode)
                     (lvalop:lval option)
                     (explist:exp list)
                     (loc:location) : unit =
  begin
    (* main() is always a state, make sure it is there *)
    if caller.ecnInfo.vname = "main" then
      (add_if states caller.ecnInfo.vname caller);
    
    (* just pretend like this call is an eel call and get its gid argument *)
    let gid_arg: eelgid_arg = (self#getGidArg (self#getGidViExpList explist)) in
    
    (* hack this so we can contruct an eelgid_operator without knowing if it is valid *)
    let getViOpName (vi_op:varinfo option) : string =
      (match vi_op with
       | Some(vi) -> vi.vname
       | _ -> "") in
    
    (* just try to get the callback func names, gives None if not 
     * there. Only use these if it is an eel call that wants the value *)
    let vi1:varinfo option = (self#getCallbackVi explist 1 loc) in
    let vi2:varinfo option = (self#getCallbackVi explist 2 loc) in
      
    (match callee.ecnInfo.vname with
    | "eel_add_timer" -> (* simple events *)
      callee.ecnLeaf <- true;
      (self#addeelCallback caller callee loc vi1 (AddTimer (loc, (getViOpName vi1), gid_arg)));
    | "eel_add_read" -> 
      callee.ecnLeaf <- true;
      (self#addeelCallback caller callee loc vi1 (AddRead (loc, (getViOpName vi1), gid_arg)));
    | "eel_add_write" -> 
      callee.ecnLeaf <- true;
      (self#addeelCallback caller callee loc vi1 (AddWrite (loc, (getViOpName vi1), gid_arg)));
    | "eel_add_error" -> 
      callee.ecnLeaf <- true;
      (self#addeelCallback caller callee loc vi1 (AddError (loc, (getViOpName vi1), gid_arg)));
    | "eel_add_read_timeout" -> (* compound events *)
      callee.ecnLeaf <- true;
      (self#addeelCallback caller callee loc vi1 (AddRead (loc, (getViOpName vi1), gid_arg)));
      (self#addeelCallback caller callee loc vi2 (AddTimeout (loc, (getViOpName vi2), gid_arg)));
    | "eel_add_write_timeout" -> 
      callee.ecnLeaf <- true;
      (self#addeelCallback caller callee loc vi1 (AddWrite (loc, (getViOpName vi1), gid_arg)));
      (self#addeelCallback caller callee loc vi2 (AddTimeout (loc, (getViOpName vi2), gid_arg)));
    | "eel_add_error_timeout" -> 
      callee.ecnLeaf <- true;
      (self#addeelCallback caller callee loc vi1 (AddError (loc, (getViOpName vi1), gid_arg)));
      (self#addeelCallback caller callee loc vi2 (AddTimeout (loc, (getViOpName vi2), gid_arg)));
    | "eel_delete_group_id" -> (* gid new/delete *)
      callee.ecnLeaf <- true;
      (add_to_HtoH caller.ecnType 
        (callee.ecnInfo.vname, callee, loc, 
          (self#getGidSource gid_arg)) (DeleteGid (loc, gid_arg)) ());
    | "eel_new_group_id" -> 
      callee.ecnLeaf <- true;
      (* get the var that is set by this call *)
      (match lvalop with 
      | Some(lexp) -> (
          let gid_var:eelgid_arg = (self#getGidArg (self#getGidVi (Lval lexp))) in
          (add_to_HtoH caller.ecnType 
            (callee.ecnInfo.vname, callee, loc, 
              (self#getGidSource gid_var)) (NewGid (loc, gid_var)) ()))
      | _ -> (fprintf stdout
                "\nWarning: eel_new_group_id called but not asigned at %s:%d\n" loc.file loc.line);)
                (* *** a call to eel_new_group_id without saving the value *)
    | _ -> ()); (* not an eel function call *)
  end

  (* just save our current fundec and create a node if we need to *)
  method vfunc (f:fundec) : fundec visitAction =
  begin
    (trace "eelstatechart" (P.dprintf "entering function %s\n" f.svar.vname));
    curFunc <- (Some (self#getNode f.svar));
    curFundec <- (Some f);
    DoChildren
  end
  
  (* just save our current stmt *)
  method vstmt (s:stmt) : stmt visitAction =
  begin
    curStmt <- (Some s);
    DoChildren
  end

  (* visit an instruction; we're only interested in calls *)
  method vinst (i:instr) : instr list visitAction =
  begin   
    curInstr <- Some(i);
    (match curFunc, curFundec, curStmt, i with
    | Some(caller), Some(f), Some(s),      (* should always be set to something *)
      Call(lvalop,e,explist,loc) -> (
        (* match calls using named functions, thereby ignoring
         * function pointers (!) *)
        match e with
        | Lval(Var(vi),NoOffset) -> (
            (* get the callee's node *)
            let callee:eelcallnode = (self#getNode vi) in
            (trace "eelstatechart" (P.dprintf "I see a call by %s to %s\n"
                                  caller.ecnInfo.vname callee.ecnInfo.vname));

            (* add one entry to each node's appropriate list *)
            (add_to_HtoQ caller.ecnCalls callee.ecnInfo.vname (callee, loc, f, s, i, explist));
            (add_to_HtoQ callee.ecnCallers caller.ecnInfo.vname (caller, loc, f, s, i, explist));
            (* check if this is an eel api call, and act accordingly *)
            (self#addeelCalls caller callee lvalop explist loc);
          )
        | _ ->
          (trace "eelstatechart" (P.dprintf "ignoring indirect call: %a\n"
                                        dn_instr i));
      )
    | _ -> ());     (* ignore other kinds of instructions *)
    DoChildren
  end
  
  (* propagate the types up the call chain using DFS *)
  method propagateTypeDFS () : unit = 
  begin   
    let updateNodeFromCall (node:eelcallnode) 
                           (callname:string) 
                           (callnode:eelcallnode)
                           (loc:location) 
                           (f:fundec)
                           (s:stmt)
                           (i:instr)
                           (explist:exp list) : unit =
    begin
      let callNodeTypeIter (src_tuple:string * eelcallnode * location * eelgid_source) 
                           (opr:eelgid_operation) (unit) : unit =
        (match src_tuple with
         call_inner, callnode_inner, loc_inner, src -> (
           (match src with
            | SrcGlobal(_)
            | SrcNew(_,_)
            | SrcUnknown
            | SrcNotVar
            | SrcCall -> (add_to_HtoH node.ecnType (callname, callnode, loc, src) opr ())
            | SrcArgument(n) -> (
              curFundec <- (Some f);
              curStmt <- (Some s);
              curInstr <- (Some i);
              let new_src = (self#getGidSource (self#getGidArg (self#getGidViExpNth explist n))) in
              (add_to_HtoH node.ecnType (callname, callnode, loc, new_src) opr ());)
           )
         )) in
      (iterHtoH callNodeTypeIter callnode.ecnType);
    end in
    (* keep a hashtable of the functions we have already visited *)
    let seen:(string,bool) Hashtbl.t = (Hashtbl.create 50) in
    (* propagate the type for this function call, recursive *)
    let rec propType (name:string) (node:eelcallnode) : unit =
    begin
      (* if we have seen it already, do nothing *)
      if (not (Hashtbl.mem seen name)) then
      begin
        (* otherwise save that we have seen it, then check all
         * of its calls *)
        (Hashtbl.add seen name true);          
        let markNodeIter (name_inner:string) 
                         (call:(eelcallnode * location * fundec * stmt * instr * exp list)) : unit = 
          (match call with
           cn, loc, f, s, i, explist -> (
             (propType name_inner cn);
             (updateNodeFromCall node name_inner cn loc f s i explist);)) in
        (iterHtoQ markNodeIter node.ecnCalls);
      end
    end in
    (* mark a state *)
    let propStateIter (name:string) (node:eelcallnode) : unit = 
      (propType name node) in
    (* start with the states, DFS to mark *)
    (* may want to make a version that does multi-pass instead of DFS to prevent stack 
     * overflow on larger programs *)
    (Hashtbl.iter propStateIter states);
    (* NOTE: this will not mark any state-stranded callback generating trees 
     * --could modify to use the whole graph is we want *)
  end
  
  (* Note to Rupak: this method was never finished *)
  (* helper function that marks gid operations in callbacks with return values *)
  (* method markReturnValues () : unit =
  begin
    let typeIter (src_tuple:string * eelcallnode * Cil.location * eelgid_source) 
                 (opr:eelgid_operation) (unit) : unit =
      (match src_tuple with 
        name, cn, loc, src -> (
          cn.
          (printDots out 1);
          (fprintf out "%s(%s:%d) %s-%s\n" name loc.file loc.line 
            (getGidSourceString src) (getGidOperationStringDetailed opr));
        )
      ) in
    
    let markState (name:string) (node:eelcallnode) : unit =
      (iterHtoH typeIter node.ecnType) in
    
    (Hashtbl.iter markState states);
  end *)
end

(* Top level helper functions *)

(* helper function to build a state chart, and callgraph *)
let computeStateChartAndGraph (f:file) : eelstates * eelcallgraph =
begin
  let scCallgraph:eelscBuildCallgraph = (new eelscBuildCallgraph) in
  (* visit the whole file, computing the call graph 
   * and callback graph *)
  (visitCilFileSameGlobals (scCallgraph :> cilVisitor) f);
  
  (* propagate the types up the callchain - verfiying gid variables along the way *)
  (scCallgraph#propagateTypeDFS ());
  
  (* mark the return values from callback functions *)
  (* (scCallgraph#markReturnValues ()); *)
   
  (* return the states *)
  ((scCallgraph#getStates ()), (scCallgraph#getGraph ()))
end

(* -------------------------------------------------------------------------- *)


(* -------------------------------------------------------------------------- *)
(* Begin State Chart output functions and classes                             *)

(* output helper functions *)

(* print space_count*2 spaces *)
let rec printSpaces (out:out_channel) (space_count:int) : unit =
begin
  if space_count > 0 then
  begin
    (fprintf out "  ");
    (printSpaces out (space_count - 1));
  end
end

(* same as printSpaces but use a dot instead of the first space *)
let rec printDots (out:out_channel) (space_count:int) : unit =
begin
  if space_count > 0 then
  begin
    (fprintf out ". ");
    (printDots out (space_count - 1));
  end
end

(* get a simple string for an eelgid_arg *)
let getGidVarString (var:eelgid_arg) : string =
  (match var with
  | ArgGlobal(vi)     -> (sprintf "Global(%s)" vi.vname)
  | ArgArgument(vi,n) -> (sprintf "Argument(%d)=%s" n vi.vname)
  | ArgLocal(vi)      -> (sprintf "Local(%s)" vi.vname)
  | ArgUnknown        -> "Unknown")
  
(* get a simple string for a location and an eelgid_arg *)
let formatGid (str:string)
              (loc:location)
              (gid_arg:eelgid_arg) : string =
  str ^ (sprintf "(%s:%d)" loc.file loc.line) ^ " " ^ (getGidVarString gid_arg)
  
(* get a simple string for a location, callback name, and an eelgid_arg *)
let formatGidCb (str:string)
                (loc:location)
                (callback_name:string) 
                (gid_arg:eelgid_arg) : string =
  (formatGid (str ^ "->" ^ callback_name) loc gid_arg)
  
(* get a detailed string for an eelgid_operation *)
let getGidOperationStringDetailed (eo:eelgid_operation) : string =
begin
  (match eo with
  | AddWrite(loc,callback_name,gid_arg) -> 
    (formatGidCb "WRITE" loc callback_name gid_arg)
  | AddRead(loc,callback_name,gid_arg) -> 
    (formatGidCb "READ" loc callback_name gid_arg)
  | AddError(loc,callback_name,gid_arg) -> 
    (formatGidCb "ERROR" loc callback_name gid_arg)
  | AddTimer(loc,callback_name,gid_arg) -> 
    (formatGidCb "TIMER" loc callback_name gid_arg)
  | AddTimeout(loc,callback_name,gid_arg) -> 
    (formatGidCb "TIMEOUT" loc callback_name gid_arg)
  | DeleteGid(loc,gid_arg) -> 
    (formatGid "DELETE" loc gid_arg)
  | NewGid(loc,gid_var) -> 
    (formatGid "NEW" loc gid_var))
end

(* get a simple string for an eelgid_operation *)
let getGidOperationStringSimple (eo:eelgid_operation) : string =
begin
  (match eo with
  | AddWrite(loc,callback_name,gid_arg) -> 
    (sprintf "WRITE -> %s" callback_name)
  | AddRead(loc,callback_name,gid_arg) -> 
    (sprintf "READ -> %s" callback_name)
  | AddError(loc,callback_name,gid_arg) -> 
    (sprintf "ERROR -> %s" callback_name)
  | AddTimer(loc,callback_name,gid_arg) -> 
    (sprintf "TIMER -> %s" callback_name)
  | AddTimeout(loc,callback_name,gid_arg) -> 
    (sprintf "TIMEOUT -> %s" callback_name)
  | DeleteGid(loc,gid_arg) -> 
    (sprintf "DELETE %s" (getGidVarString gid_arg))
  | NewGid(loc,gid_var) -> 
    "NEW")
end

(* get a DOT formatted string for an eelgid_operation *)
let getGidOperationStringDot (eo:eelgid_operation) : string = 
begin
  (match eo with
  | AddWrite(loc,callback_name,gid_arg) -> 
    (sprintf "%s [label=\"W\"]" callback_name)
  | AddRead(loc,callback_name,gid_arg) -> 
    (sprintf "%s [label=\"R\"]" callback_name)
  | AddError(loc,callback_name,gid_arg) -> 
    (sprintf "%s [label=\"E\"]" callback_name)
  | AddTimer(loc,callback_name,gid_arg) -> 
    (sprintf "%s [label=\"T\"]" callback_name)
  | AddTimeout(loc,callback_name,gid_arg) -> 
    (sprintf "%s [label=\"TO\",color=grey,fontcolor=grey]" callback_name)
  | DeleteGid(loc,gid_arg) -> 
    "delete [color=grey,fontcolor=grey]"
  | NewGid(loc,gid_var) -> 
    "new")
end

(* get a simple string for an eelgid_source *)
let getGidSourceString (es:eelgid_source) :  string =
begin
  (match es with
    | SrcGlobal(vi)  -> (sprintf "Global(%s)" vi.vname)
    | SrcArgument(n) -> (sprintf "Argument(%d)" n)
    | SrcNew(vi,loc) -> (sprintf "New(%s@%s:%d)" vi.vname loc.file loc.line)
    | SrcUnknown     -> "Unknown"
    | SrcNotVar      -> "NotVar"
    | SrcCall        -> "Call")
end

(* get a simple string for an eelgid_operation trasition *)
let getTransitionOnlyString (eo:eelgid_operation) : string = 
begin
  (match eo with
  | AddWrite(loc,callback_name,gid_arg) -> 
    (sprintf "(%s)W" callback_name)
  | AddRead(loc,callback_name,gid_arg) -> 
    (sprintf "(%s)R" callback_name)
  | AddError(loc,callback_name,gid_arg) -> 
    (sprintf "(%s)E" callback_name)
  | AddTimer(loc,callback_name,gid_arg) -> 
    (sprintf "(%s)T" callback_name)
  | AddTimeout(loc,callback_name,gid_arg) -> 
    (sprintf "(%s)TO" callback_name)
  | DeleteGid(loc,gid_arg) -> 
    "delete"
  | NewGid(loc,gid_var) -> 
    "new")
end

(* helper function that dumps all possible gid operations in a verbose manner *)
let printFullTypeGraph (out:out_channel) (states:eelstates) : unit =
begin
  let printEventTypeString (h:eelTypeHash) : unit =
  begin
    let hIter (src_tuple:string * eelcallnode * Cil.location * eelgid_source) 
              (opr:eelgid_operation) (unit) : unit =
    begin
      (match src_tuple with 
        name, cn, loc, src -> (
          (printDots out 1);
          (fprintf out "%s(%s:%d) %s-%s\n" name loc.file loc.line 
            (getGidSourceString src) (getGidOperationStringDetailed opr));
        )
      );
    end in
    (iterHtoH hIter h);
  end in
  
  let printState (name:string) (node:eelcallnode) : unit =
  begin
    (fprintf out "\n{State %s}\n" name);
    (printEventTypeString node.ecnType);
  end in
  (fprintf out "\n{FULL STATE TYPE DUMP}\n");
  (sortByKeyAndIter printState states);
  (fprintf out "\n{/FULL STATE TYPE DUMP}\n");
end

(* eelstatechart visitors - traverse the chart *)

(* eelstatechart visitor type *)
class type scVisitor = object
  method beginChart: int -> string -> eelgid_source -> unit
  method transition: string -> eelgid_operation -> unit
  method state: string -> int list -> unit
  method endChart: unit -> unit
  method error: string -> unit
end

(* no op visitor *)
class nopScVisitor : scVisitor = object
  method beginChart (chart_num:int) (start_state:string) (lookfor:eelgid_source) : unit = ()
  method transition (from_state:string) (opr:eelgid_operation) : unit = ()
  method state (state_name:string) (gens:int list) : unit = ()
  method endChart (unit) : unit = ()
  method error (str:string) : unit = ()
end

(* dot chart visitor *)
class dotChartVisitor(out:out_channel) = object(self) 
  inherit nopScVisitor 
  
  (* skip redundant transitions *)
  val seen_transitions: (string * string, bool) Hashtbl.t = (Hashtbl.create 10)
  (* have we seen a transition to delete? (because you do not get a "state" call for delete) *)
  val mutable seen_delete: bool = false
  (* save the states to output at end (so we can figure out who has a TO pointed) *)
  val states: (string, bool * int list) Hashtbl.t = (Hashtbl.create 5)
  
  (* Note to Rupak - charts are generated all in a single dot file -> a single ps file
   * the ps file always has a page sixe to match the first chart which is sometimes
   * smaller than later charts (plb). It might make sense to change this to produce a
   * separate dot file = separate ps for each chart. *)
  
  (* start of a chart *)
  method beginChart (chart_num:int) (start_state:string) (lookfor:eelgid_source) : unit = 
  begin
    (Hashtbl.clear seen_transitions);
    seen_delete <- false;
    (Hashtbl.clear states);
    (fprintf out 
      ("\ndigraph g%d {\n" ^^
(*      "  page=\"8.5,11\";\n" ^^*)   (* these do not seem to help the page sizing problem *)
(*      "  size=\"7.9,10.6\";\n" ^^*)
      "  size=\"7.5,10\";\n" ^^
      "  rotate=90;\n" ^^
      "  label=\"Chart g%d: %s - %s\"\n" ^^
      "  node [shape=box,width=.1,height=.1];\n")
      chart_num
      chart_num
      start_state
      (getGidSourceString lookfor));
  end
  
  (* saw a transition *)
  method transition (from_state:string) (opr:eelgid_operation) : unit = 
  begin
    (* make sure we do not show a duplicate transition
     * i.e. same exact transition generated, but from the different 
     * call location *)
    let to_state:string = (getTransitionOnlyString opr) in
    if (not (Hashtbl.mem seen_transitions (from_state, to_state))) then
    begin
      (Hashtbl.add seen_transitions (from_state, to_state) true);
      (fprintf out "  %s -> %s;\n" from_state (getGidOperationStringDot opr));
      (match opr with
      | DeleteGid(_,_) when (not seen_delete) -> (
        seen_delete <- true;
        (fprintf out "  delete [color=grey,fontcolor=grey];\n");)
      | AddTimeout(_,callback_name,_) -> (
        if (Hashtbl.mem states callback_name) then
        begin
          let old = (Hashtbl.find states callback_name) in
          (Hashtbl.remove states callback_name);
          (Hashtbl.add states callback_name (true, (snd old)));
        end  
        else
          (Hashtbl.add states callback_name (true, [])))
      | _ -> ());
    end
  end
  
  (* saw a state *)
  method state (state_name:string) (gens:int list) : unit = 
  begin
    if (Hashtbl.mem states state_name) then
    begin
      let old = (Hashtbl.find states state_name) in
      (Hashtbl.remove states state_name);
      (Hashtbl.add states state_name ((fst old), gens));
    end
    else
      (Hashtbl.add states state_name (false, gens));
  end
  
  (* end of chart *)
  method endChart (unit) : unit = 
  begin
    let stateIter (state_name:string) (grey_and_gens:bool * int list) : unit =
    begin
      let grey:bool = (fst grey_and_gens) in
      let gens:int list = (snd grey_and_gens) in
      if (List.length gens) > 0 then
      begin
        let foldGens (s:string) (n:int) : string =
          (s ^ (sprintf "-g%d" n)) in
        (fprintf out 
          "  %s [label=\"%s %s\"%s];\n"
          state_name
          state_name
          (List.fold_left foldGens "" gens)
          (if grey then 
             ",color=grey,fontcolor=grey" 
           else 
             ""));
      end
      else
      begin
        if grey then 
          (fprintf out "  %s [color=grey,fontcolor=grey];\n" state_name);
      end
    end in
    (Hashtbl.iter stateIter states);
    (fprintf out "}\n");
  end
  
  (* an error happened during the visit *)
  method error (str:string) : unit = 
    (fprintf out "\n\"erorr %s\";\n" str)
end

(* simple text chart output *)
class simpleChartVisitor(out:out_channel) = object 
  inherit nopScVisitor 
  
  method beginChart (chart_num:int) (start_state:string) (lookfor:eelgid_source) : unit = 
    (fprintf out "\n{chart %d %s - %s}\n" chart_num start_state (getGidSourceString lookfor))
  method transition (from_state:string) (opr:eelgid_operation) : unit =
    (fprintf out "%s:%s\n" from_state (getGidOperationStringDetailed opr))
  method state (state_name:string) (gens:int list) : unit =
  begin
    (fprintf out "{state %s " state_name);
    (List.iter (fun y -> (fprintf out "%d " y)) gens);
    (fprintf out "}\n"); 
  end
  method endChart (unit) : unit = 
    (fprintf out "{end chart}\n")
  method error (str:string) : unit = 
    (fprintf out "\nerror=%s\n" str)
end

(* the big function to visit a statechart *)
let visitStateChart (states:eelstates) (vis:scVisitor) : unit =
begin
  (* save the calls to eel_new_group_id - they each create a chart *)
  let news: (eelgid_source, int * string) Hashtbl.t = (Hashtbl.create 1) in
  (* save what charts a particular state can generate ("-g#" in chart) *)
  let gen_chart: (string, (int, bool) Hashtbl.t) Hashtbl.t = (Hashtbl.create 5) in
  
  (* get a list of a hashtable's keys (should be in general helpers section) *)
  let hashToFirstList (table:('a, 'b) Hashtbl.t) : 'a list =
  begin
    let foldFunc 
            (key:'a) 
            (value:'b) 
            (l:'a list) : 'a list =
      key :: l in
    (Hashtbl.fold foldFunc table [])
  end in
  
  (* pull out the chart numbers from the gen charts information *)
  let getChartGens (state:string) : int list =
    try
      (List.sort compare (hashToFirstList (Hashtbl.find gen_chart state)))
    with _ -> []
    in
  
  (* do the group ids equal each other *)
  let equalSrc (s1:eelgid_source) (s2:eelgid_source) : bool =
    (match s1, s2 with
    | SrcGlobal(vi1), SrcGlobal(vi2) -> (vi1.vid = vi2.vid)
    | SrcNew(vi1,loc1), SrcNew(vi2,loc2) -> ((vi1.vid = vi2.vid) && (loc1 = loc2))
    | _,_ -> (s1 = s2)
    ) in
  
  (* the real chart visit function, but just calls up handle_state function *)
  let visitChart (start_state:string) (lookfor:eelgid_source) (handle_state) : unit =
  begin
    let statesToDo:(string * eelgid_source) Queue.t = (Queue.create ()) in
    (Queue.add (start_state, lookfor) statesToDo);
    
    let iterStateTransitions (name:string) (node:eelcallnode) (src:eelgid_source) : unit = 
    begin
      (* this is a new state *)
      let handle_type = (handle_state name node src) in
      let stateTypeIter (src_tuple:string * eelcallnode * Cil.location * eelgid_source) 
                        (opr:eelgid_operation) (unit) : unit =
      begin
        (* check this opr, only interested in oprs on src *)
        (match src_tuple with 
        | (name, cn, loc, s) when (equalSrc s src) ->
            (match opr with 
            | AddWrite(loc,callback_name,gid_arg)
            | AddRead(loc,callback_name,gid_arg)
            | AddError(loc,callback_name,gid_arg)
            | AddTimer(loc,callback_name,gid_arg)
            | AddTimeout(loc,callback_name,gid_arg) -> 
              (Queue.add (callback_name, (SrcArgument 0)) statesToDo)
            | DeleteGid(loc,gid_arg) -> ()
            | NewGid(loc,gid_var) -> ()
            )
        | _,_,_,_ -> ()
        );
        (handle_type src_tuple opr);
      end in      
      (* iterate through type infd for the state *)
      (iterHtoH stateTypeIter node.ecnType);
    end in
    (* do the other states until done *)
    let seenState:(string, bool) Hashtbl.t = (Hashtbl.create 10) in
    let rec iterQ (unit) : unit = 
    begin
      if (not (Queue.is_empty statesToDo)) then
      begin
        let name_pair:string * eelgid_source = (Queue.pop statesToDo) in
        if (not (Hashtbl.mem seenState (fst name_pair))) then
        begin
          try
            (Hashtbl.add seenState (fst name_pair) true);
            let node:eelcallnode = (Hashtbl.find states (fst name_pair)) in
            (iterStateTransitions (fst name_pair) node (snd name_pair));
            (iterQ ());
          with _ -> (
            (vis#error (sprintf "Can't find 1 state %s" (fst name_pair)));
            (iterQ ());
            )
        end
        else
        begin
          (iterQ ());
        end
      end
    end in
    (iterQ ());
  end in
  
  (* find the eel_new_group_id calls *)
  let findNews (start_state:string) : unit =
  begin
    let current_chart_number: mutable_int = {value_int=0} in
    (* get news from a state *)
    let findNewsInState (name:string) (state:eelcallnode) : unit =
    begin
      let iterNewStateType (src_tuple:string * eelcallnode * Cil.location * eelgid_source) 
                           (opr:eelgid_operation) (unit) : unit =
      begin
        let addNew (src:eelgid_source) : unit =
        begin
          (Hashtbl.add news src (current_chart_number.value_int, name));
          (add_to_HtoH gen_chart name current_chart_number.value_int true);
          current_chart_number.value_int <- current_chart_number.value_int + 1;
        end in
        (match src_tuple with 
        | (iname, cn, loc, SrcNew(vi, iloc)) -> (
          if (Hashtbl.mem news (SrcNew (vi, iloc))) then
            (add_to_HtoH gen_chart name (fst (Hashtbl.find news (SrcNew (vi, iloc)))) true)
          else
            (addNew (SrcNew (vi, iloc)))
          )
        | _,_,_,_ -> ());
        (* now also check for new global ids *)
        (match opr with
        | NewGid(iloc,ArgGlobal(gvi)) -> (addNew (SrcGlobal gvi))
        | _ -> ()
        );
      end in
      (* sort type to show news first *)
      (iterTypeByNews iterNewStateType state.ecnType);
    end in
    (* just get the news from start state - do this one first *)
    if (not (Hashtbl.mem states start_state)) then
    begin
      (vis#error (sprintf "Can't find 2 state %s" start_state));
    end
    else
    begin
      (findNewsInState start_state (Hashtbl.find states start_state));  
      (* now do all the others *)    
      let iterTopStates (name:string) (state:eelcallnode) : unit =
        if (not (name = start_state)) then
          (findNewsInState name state) in
      (Hashtbl.iter iterTopStates states);
    end
  end in
  
  (* visit the chart calling the visitor's methods *)
  let finalVisitChart (chart_number:int) (start_state:string) (lookfor:eelgid_source) : unit =
  begin
    (vis#beginChart chart_number start_state lookfor);
        
    let handle_state (name:string) (node:eelcallnode) (src:eelgid_source) = 
    begin
      let seen:(eelgid_operation, bool) Hashtbl.t = (Hashtbl.create 3) in
      let handleOpr (opr:eelgid_operation) : unit =
      begin
        if (not (Hashtbl.mem seen opr)) then
        begin
          (Hashtbl.add seen opr true);
          (vis#transition name opr);
        end
      end in
      let handle_type (src_tuple:string * eelcallnode * Cil.location * eelgid_source) 
                      (opr:eelgid_operation) : unit =
      begin
        (* check this chart's opr, only interested in oprs on src *)
        (match src_tuple with 
        | (name, cn, loc, s) when (equalSrc s src) ->
            (match opr with 
            | AddWrite(_,_,_)
            | AddRead(_,_,_)
            | AddError(_,_,_)
            | AddTimer(_,_,_)
            | AddTimeout(_,_,_)
            | DeleteGid(_,_) -> 
              (handleOpr opr)
            | NewGid(_,_) -> ()
            )
        | _,_,_,_ -> ()
        );
      end in
      (* this is a new state *)
      (vis#state name (getChartGens name));
      (* return our func to handle the type *)
      handle_type
    end in
    
    (* do the actual visit *)
    (visitChart start_state lookfor handle_state);
  
    (* mark the end of the chart *)
    (vis#endChart ());
  end in
  
  (* find all the non-global news *)
  (findNews "main");
  
  (* do charts for each global News, sort by chart number first *)
  let iterNews (tr:int * string * eelgid_source) : unit =
    (match tr with a,b,c -> (finalVisitChart a b c)) in
  let compareFirstOfThree (tr1:int * string * eelgid_source) (tr2:int * string * eelgid_source) : int =
    (match tr1,tr2 with (n1,_,_),(n2,_,_) -> (compare n1 n2)) in
  let foldNews (src:eelgid_source) (p:(int * string)) 
      (l:(int * string * eelgid_source) list) : (int * string * eelgid_source) list =
    (match p with num,name -> (num,name,src) :: l) in
  (List.iter iterNews (List.sort compareFirstOfThree (Hashtbl.fold foldNews news [])));
end

(* -------------------------------------------------------------------------- *)


(* -------------------------------------------------------------------------- *)
(* Begin BLAST classes                                                        *)

(* BLAST translation state chart visitor for independent state verification *)
class blastIndependentChartVisitor() = object(self) 
  inherit nopScVisitor
  
  val mutable gid_init:string = ""
  val mutable gid_name:string = ""
  val mutable test_function_text:string = ("\nvoid *new_void_NONDET__SCG();\n" ^ 
                                           "int new_fd_NONDET__SCG();\n" ^ 
                                           "void check_gid_properties__SCG(eel_group_id gid);\n\n" ^ 
                                           "int independent_state_check_entry__SCG(int argc , char **argv ) \n{\n")
  (*val mutable test_function_text:string = ("\n\nint independent_state_check_entry__SCG(int argc , char **argv ) \n{\n") *)
                                           
  method getTestFunctionText () : string =
    (test_function_text ^ "\n  exit(0);\n  return 0;\n}\n")

  method beginChart (chart_num:int) (start_state:string) (lookfor:eelgid_source) : unit = 
    (match lookfor with
    | SrcGlobal(vi) -> (
      gid_init <- vi.vname;
      gid_name <- vi.vname;
      test_function_text <- (test_function_text ^ 
        (sprintf "\n  // testing GLOBAL gid %s\n  {\n" vi.vname));)
    | SrcNew(vi,loc) -> (
      gid_init <- "eel_group_id gid";
      gid_name <- "gid";
      test_function_text <- (test_function_text ^ 
        (sprintf "\n  // testing NEW gid %s@%s:%d\n  {\n" vi.vname loc.file loc.line));)
    | _ -> (self#error "Unknown chart type"))
    
  method transition (from_state:string) (opr:eelgid_operation) : unit =
    ()
    
  method state (state_name:string) (gens:int list) : unit =
    if state_name = "main" then () else
      (test_function_text <- 
        (test_function_text ^ 
        (sprintf ("\n    {" ^^
                  "\n      %s = eel_new_group_id();" ^^ 
                  "\n      %s(%s,new_void_NONDET__SCG(),new_fd_NONDET__SCG());" ^^ 
                  "\n      check_gid_properties__SCG(%s);" ^^ 
                  "\n    }\n")
                  gid_init state_name gid_name gid_name));)
    
  method endChart (unit) : unit = 
    (test_function_text <- (test_function_text ^ "  }\n");)
    
  method error (str:string) : unit = 
    (fprintf stdout "\nError while visiting state chart for BLAST independent translation=%s\n" str) 
end


(* cil visitor that does all sorts of stuff including
 * function definition replacements and function call replacements.
 * It has some (in progress) code for doing full program transformation. *)
(* translate a eel file to a BLAST file *)
class eelToBlast (states:eelstates) (graph:eelcallgraph) 
                 (ndH:(string,bool) Hashtbl.t) (retsH:(string,int) Hashtbl.t) 
                 (fulltrans:bool) = object(self)
  inherit nopCilVisitor
  
  (* helper bools to switch modes when visiting cil file *)
  val mutable doglobals:bool = false
  val mutable docalls:bool = false
  
  (* general replacements (can be user or scg) *)  
  val replace_call_vars:(string * location option, varinfo * exp list) Hashtbl.t = (Hashtbl.create 10)
  val add_vars:(string, varinfo Queue.t) Hashtbl.t = (Hashtbl.create 5)
  val mutable add_globals:global list = []
  
  (* eel interface information *)
  val eel_vis:(string, varinfo) Hashtbl.t = (Hashtbl.create 10) 
  val mutable group_count:int = 0
  val mutable max_events_per_group:int = 0
  val global_state_vars:(string, varinfo) Hashtbl.t = (Hashtbl.create 5)
  
  (* these are used for state name->number conversion *)
  val state_number:(string,int*varinfo) Hashtbl.t = (Hashtbl.create 5)
  val number_state:(int,string*varinfo) Hashtbl.t = (Hashtbl.create 5)
  
  (* helper functions  that build cil types *)
  method createBLASTvi (vi_in:varinfo) : varinfo =
    { (copyVarinfo vi_in (vi_in.vname ^ "__SCG")) with vstorage = NoStorage; }
  method createBLASTviUser (vi_in:varinfo) : varinfo =
    { (copyVarinfo vi_in (vi_in.vname ^ "__SCG_USER")) with vstorage = NoStorage; }
  
  method createBLASTRetIntfundec (new_vi:varinfo) (ret_value:int) : fundec =
  begin
    let getStmts (vii:varinfo) : stmt list =
      (* just return ret_value if a return is required *)
      let ret:stmt = (mkStmt (Return ((Some (integer ret_value)), locUnknown))) in
      (match vii.vtype with
      | TFun(t,_,_,_) -> 
        (match t with 
        | TVoid(_) -> []
        | _ -> [ret])
      | _ -> []) in
      
    (self#createBLASTfundec new_vi (getStmts new_vi))
  end
  
  method createBLASTfundec (new_vi:varinfo) (stmts:stmt list) : fundec =
  begin
    let b:block = (mkBlock stmts) in
    let f:fundec = {
      svar = new_vi;
      sformals = [];
      slocals = [];
      smaxid = 0;
      sbody = b;
      smaxstmtid = None;
      sallstmts = [];
    } in
    (* ignore the formals for now since we do not have the varinfos of them,
     * just name and types *)
    (setMaxId f);
    f
  end
  
  method replaceFuncPtrTypes (vi:varinfo) : unit =
  begin
    let matchFunc (arg) =
      (match arg with 
      | name, TPtr(TFun(_,_,_,_),_), _ -> (name, (TInt (IInt,[])), [])
      | _ -> arg) in
    let modType (t:typ) : typ =
      (match t with
      | TFun(l,Some(args),b,a) -> (TFun (l,(Some (List.map matchFunc args)),b,a))
      | _ -> t) in
    vi.vtype <- (modType vi.vtype)
  end
  
  method mapFuncPtrs (el:exp list) : exp list =
  begin
    let replacePtr (e) : exp = 
     (match e with
     | AddrOf(l)
     | Lval(l) ->
       (match l with 
       | Var(vi),NoOffset ->
         (match vi.vtype with
         | TFun(_,_,_,_) -> (integer (fst (Hashtbl.find state_number vi.vname)))
         | _ -> e)
       | _ -> e)
     | _ -> e) in
    (List.map replacePtr el)
  end

  (* run the BLAST class over a cil file *)
  method run (f:file) : unit =
  begin 
    try
      if fulltrans then
      begin
        (* This creates all the necessarily vis and fundecs for full program translation *)
        (* get all the calls to new gid *)
        let eel_new_gid:string = "eel_new_group_id" in
        let eel_new_cn:eelcallnode = (Hashtbl.find graph eel_new_gid) in
        
        (* add the new gid prototype *)
        let new_gid_vi:varinfo = (self#createBLASTvi eel_new_cn.ecnInfo) in
        new_gid_vi.vtype <- (
          match new_gid_vi.vtype with
          | TFun(rt,_,_,_) -> (TFun (rt, (Some [("group_number", (TInt (IInt,[])), [])]), false, []))
          | _ -> new_gid_vi.vtype);
        (add_to_HtoQ add_vars eel_new_gid new_gid_vi); 
        
        (* replace all calls to new gid with the new gid call *)
        let iterNewCalls (func_name:string) 
          (call_details:eelcallnode * location * fundec * stmt * instr * exp list) : unit =
          (match call_details with
          | ecn,loc,fd,s,i,el -> (
            group_count <- group_count + 1;
            (add_if replace_call_vars (eel_new_gid,(Some loc)) 
              (new_gid_vi,[(integer group_count)]));
            )
          ) in
        (iterHtoQ iterNewCalls eel_new_cn.ecnCallers); 
        
        (* replace eel_dispatch calls *)
        let replaceCallsAndProto (name:string) =
        begin
          let cn:eelcallnode = (Hashtbl.find graph name) in
          let vi:varinfo = (self#createBLASTvi cn.ecnInfo) in
          (add_to_HtoQ add_vars name vi); 
          (add_if replace_call_vars (name,None) (vi,[]));
        end in
        (replaceCallsAndProto "eel_dispatch_loop");
        (replaceCallsAndProto "eel_delete_group_id");
        (replaceCallsAndProto "eel_remove_events");
        
        (* number states and count max events per group *)
        let state_count = ref 0 in
        let iterState (state_name:string) (state:eelcallnode) : unit =
        begin
          state_count := !state_count + 1;
          (add_if state_number state_name (!state_count,state.ecnInfo));
          (add_if number_state !state_count (state_name,state.ecnInfo));
          let max_arg_events = ref 0 in
          let iterType (src_tuple:string * eelcallnode * Cil.location * eelgid_source) 
                      (opr:eelgid_operation) (unit)  : unit =
            (match src_tuple with cbname,_,_,src -> 
              (match src with 
              | SrcGlobal(_) -> () (* ignore global for now *)
              | SrcArgument(n) when n = 0 -> (max_arg_events := !max_arg_events + 1)
              | _ -> ())) in
          (iterHtoH iterType state.ecnType);
          if (!max_arg_events > max_events_per_group) then
            max_events_per_group <- !max_arg_events;
        end in
        (Hashtbl.iter iterState states);
        
        (*(visitStateChart states ((new blastChartVisitor ...) :> scVisitor));*)
      end
      else
      begin
        (* This creates a stub function that can be used for verifying states independently *)
        
        (* just cheat and use the comment text to create a new test __SCG function *)
        let visBLAST:blastIndependentChartVisitor = (new blastIndependentChartVisitor ()) in
        (visitStateChart states (visBLAST :> scVisitor));
        add_globals <- (GText (visBLAST#getTestFunctionText ())) :: add_globals;
      end;
      
      (* do all the global replacements first *)
      doglobals <- true;
      (visitCilFile (self :> cilVisitor) f);
      doglobals <- false;
      (* otherwise we will miss some replacements (not sure why cil works that way: 
       * seems like visit should allow global changes, otherwise there would not be a "SameGlobals" version) *)
      docalls <- true;
      (visitCilFileSameGlobals (self :> cilVisitor) f);
      docalls <- false;
      (* finally, just add any globals (that you did not want visited) to the end of the file *)
      f.globals <- (List.concat [f.globals; add_globals]);
      
    with _ -> (fprintf stdout "Error while running eelToBlast transform.\n"); 
  end
  
  (* cil visitor global variable handler *)
  method vglob (g:global) = 
  begin
    if doglobals then 
    begin
      (* collect the types we need *)
      (try
        (match g with 
        (* add a new prototype after the orignal, the new one will not be implemented
         * which will cause BLAST to see it as non-deterministic *)
        | GVarDecl(vi,loc) when (Hashtbl.mem ndH vi.vname) -> (
          let vi_new:varinfo = (self#createBLASTviUser vi) in
          (add_if replace_call_vars (vi.vname,None) (vi_new,[]));
          (ChangeTo [g; (GVarDecl (vi_new, locUnknown))])
          )
        (* after the prototype of a func add a definition of our replacement func 
         * NOTE: if no prototype, this will not work...but would need to share
         * varinfos if we just put a prototype here and definition later *)
        | GVarDecl(vi,loc) when (Hashtbl.mem retsH vi.vname) -> (
          let f:fundec = (self#createBLASTRetIntfundec (self#createBLASTviUser vi) (Hashtbl.find retsH vi.vname)) in
          (add_if replace_call_vars (vi.vname,None) (f.svar,[]));
          (ChangeTo [g; (GFun (f, locUnknown))])
          )
        (* once you see a certain prototype, add in some new prototypes 
         * NOTE: if there is no prototype this will fail to add the new protos *)
        | GVarDecl(vi,loc) when (Hashtbl.mem add_vars vi.vname) -> (
          let vQ:varinfo Queue.t = (Hashtbl.find add_vars vi.vname) in
          let foldQ (gl:global list) (vi:varinfo) : global list =
            ((GVarDecl (vi, locUnknown)) :: gl) in
          (ChangeTo (g :: (Queue.fold foldQ [] vQ)))
          )
        | GFun(f,loc) when (fulltrans && (Hashtbl.mem state_number f.svar.vname)) -> 
          (ChangeTo [(GText (sprintf "// STATE number %d" (fst (Hashtbl.find state_number f.svar.vname)))); g])
        (* add all the fake eel prototypes, put at bottom because String.sub could throw an exception *)
        | GVarDecl(vi,loc) when (fulltrans && ((String.sub vi.vname 0 7) = "eel_add")) -> (
          let vi_new:varinfo = (self#createBLASTvi vi) in
          (self#replaceFuncPtrTypes vi_new);
          (add_if eel_vis vi.vname vi_new);
          (ChangeTo [g; (GVarDecl (vi_new, locUnknown))]))
        
        | _ -> DoChildren)
      with _ -> DoChildren)
    end
    else
      DoChildren
  end
  
  (* cil visitor instr handler *)
  method vinst (i:instr) = 
  begin
    if docalls then
    begin
      (match i with
      (* replace any calls to our stubbed functions, by location *)
      | Call(lv,Lval(Var(vi),NoOffset),args,loc) when (Hashtbl.mem replace_call_vars (vi.vname,(Some loc))) -> 
        (match (Hashtbl.find replace_call_vars (vi.vname,(Some loc))) with vi_new,args_add -> 
          (ChangeTo [(Call (lv,Lval((Var vi_new),NoOffset),(List.concat [args_add; args]),loc))]))
      (* replace any calls to our stubbed functions, at any location *)
      | Call(lv,Lval(Var(vi),NoOffset),args,loc) when (Hashtbl.mem replace_call_vars (vi.vname,None)) -> 
        (match (Hashtbl.find replace_call_vars (vi.vname,None)) with vi_new,args_add ->
          (ChangeTo [(Call (lv,Lval((Var vi_new),NoOffset),(List.concat [args_add; args]),loc))]))
      (* eel calls that need state name->number parameter conversion *)    
      | Call(lv,Lval(Var(vi),NoOffset),args,loc) when (Hashtbl.mem eel_vis vi.vname) -> 
        (let vi_new:varinfo = (Hashtbl.find eel_vis vi.vname) in
          (ChangeTo [(Call (lv,Lval((Var vi_new),NoOffset),(self#mapFuncPtrs args),loc))]))
          
      | _ -> DoChildren)
    end
    else
      DoChildren
  end

end

(* top level function for doing BLAST transformations *)
let translateToBlast (f:file) (states:eelstates) (graph:eelcallgraph) 
                     (blastnd:string) (blastrets:string) (fulltrans:bool) : unit =
begin
  let ndH: (string, bool) Hashtbl.t = (Hashtbl.create 1) in
  let retsH: (string, int) Hashtbl.t = (Hashtbl.create 1) in
  (try
    let iternd (s:string) : unit =
      (add_if ndH s true) in
    let iterrets (s:string) : unit =
      (Scanf.sscanf s "%s@(%d)" (fun s n -> (add_if retsH s n))) in
    (List.iter iternd (Str.split (Str.regexp ",") blastnd));
    (List.iter iterrets (Str.split (Str.regexp ",") blastrets));
  with _ -> (fprintf stdout "Error while parsing options.\n"));
  try
    let vis:eelToBlast = (new eelToBlast states graph ndH retsH fulltrans) in
    (vis#run f);
  with _ -> (fprintf stdout "Error while translating file to BLAST input.\n");
end

(* -------------------------------------------------------------------------- *)


(* -------------------------------------------------------------------------- *)
(* Begin cil module implementation                                            *)

(* boolean reference that indicates if we should run *)
let doeelstatechart = ref false
let blastOut = ref false
let blastNondet = ref ""
let blastRets = ref ""
let blastFullTrans = ref false

(* our feature description instance *)
let feature : featureDescr = 
  { fd_name = "eelstatechart";
    fd_enabled = doeelstatechart;
    fd_description = "generation of an eel state chart";
    fd_extraopt = [
      ("--blast-out", Arg.Unit (fun _ -> blastOut := true),
           "translate the program into BLAST input");
      ("--blast-nondet", Arg.String (fun s -> blastNondet := s),
           "=<s>: substitute non-deterministic function prototypes for function names separated by ,");
      ("--blast-rets", Arg.String (fun s -> blastRets := s),
           "=<s>: substitute functions returning an int n for fun(n) separated by ,"); 
      ("--blast-full-trans", Arg.Unit (fun _ -> blastFullTrans := true),
           "translate the full program (eel_dispatch__SCG), otherwise creates independent state main__SCG"); ];
    fd_doit = 
    (function (f: file) -> 
      (* this gets rid of a bunch of unused labels, etc *)
      (Rmtmps.removeUnusedTemps f);
      (* compute the state chart *)
      let sByg:(eelstates * eelcallgraph) = (computeStateChartAndGraph f) in
      let states:eelstates = (fst sByg) in
      let graph:eelcallgraph = (snd sByg) in
      
      (* these are both really just for debug *)
      (printFullTypeGraph stdout states);
      (visitStateChart states ((new simpleChartVisitor stdout) :> scVisitor));
      (* print the state chart(s) out in dot format *)
      (try
        let dotfile:out_channel = (open_out "eelstatechart.dot.txt") in
        (visitStateChart states ((new dotChartVisitor dotfile) :> scVisitor));
        (close_out dotfile);
      with _ -> (fprintf stdout "Failed to open dot file for writing.\n"));
      
      (* convert the file to blast input if requested *)
      if !blastOut then
        (translateToBlast f states graph !blastNondet !blastRets !blastFullTrans);
    );
    fd_post_check = false;
  } 

(* -------------------------------------------------------------------------- *)
